import numpy
from typing import overload

CAN_THREAD: bool
GENERATOR: RandomGenerator

class AbstractPulseTrain:
    n_used_electrodes: int
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def n_delta_t(self) -> int: ...
    @property
    def n_electrodes(self) -> int: ...
    @property
    def n_pulses(self) -> int: ...
    @property
    def n_unique_pulses(self) -> int: ...
    @property
    def sigma_ap(self) -> float: ...
    @property
    def steps_to_ap(self) -> int: ...
    @property
    def t_max(self) -> int: ...
    @property
    def time_step(self) -> float: ...
    @property
    def time_to_ap(self) -> float: ...

class ConstantPulseTrain(AbstractPulseTrain):
    def __init__(self, duration: float, rate: float, amplitude: float, time_step: float = ..., time_to_ap: float = ..., sigma_ap: float = ...) -> None: ...
    def get_pulse(self, arg0: int) -> Pulse: ...
    @property
    def amplitude(self) -> float: ...
    @property
    def pulse_interval(self) -> int: ...

class Decay:
    def __init__(self, *args, **kwargs) -> None: ...
    def compute_accommodation(self, arg0: int, arg1: FiberStats) -> float: ...
    def compute_adaptation(self, arg0: int, arg1: FiberStats, arg2: list[float]) -> float: ...
    def decay(self, arg0: int) -> float: ...
    def randomize(self, arg0: RandomGenerator) -> Decay: ...
    @property
    def time_step(self) -> float: ...

class Exponential(HistoricalDecay):
    def __init__(self, adaptation_amplitude: float = ..., accommodation_amplitude: float = ..., 
                 sigma_adaptation_amplitude: float = ..., sigma_accommodation_amplitude: float = ..., 
                 exponents: list[tuple[float, float]] = ...,
                 memory_size: int = ..., allow_precomputed_accommodation: bool = ..., cached_decay: bool = ..., cache: list[float] = ...) -> None: ...
    @property
    def exponents(self) -> list[tuple[float, float]]: ...

class Fiber:
    i_det: list[float]
    sigma: list[float]
    spatial_constant: list[float]
    def __init__(self, i_det: list[float], spatial_constant: list[float], sigma: list[float], fiber_id: int, sigma_rs: float = ..., refractory_period: RefractoryPeriod = ..., rel, decay: Decay = ..., store_stats: bool = ...) -> None: ...
    def process_pulse(self, arg0: Pulse, arg1: AbstractPulseTrain) -> None: ...
    def randomize(self) -> Fiber: ...
    @property
    def decay(self) -> Decay: ...
    @property
    def refractory_period(self) -> RefractoryPeriod: ...
    @property
    def stats(self) -> FiberStats: ...
    @property
    def threshold(self) -> float: ...

class FiberStats:
    def __init__(self, fiber_id: int = ...) -> None: ...
    def __eq__(self, arg0: FiberStats) -> bool: ...
    @property
    def accommodation(self) -> numpy.ndarray: ...
    @property
    def adaptation(self) -> numpy.ndarray: ...
    @property
    def fiber_id(self) -> int: ...
    @property
    def n_pulses(self) -> int: ...
    @property
    def n_spikes(self) -> int: ...
    @property
    def pulse_times(self) -> numpy.ndarray: ...
    @property
    def refractoriness(self) -> numpy.ndarray: ...
    @property
    def scaled_i_given(self) -> numpy.ndarray: ...
    @property
    def spikes(self) -> numpy.ndarray: ...
    @property
    def stochastic_threshold(self) -> numpy.ndarray: ...
    @property
    def trial_id(self) -> int: ...

class HistoricalDecay(Decay):
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def accommodation_amplitude(self) -> float: ...
    @property
    def adaptation_amplitude(self) -> float: ...
    @property
    def memory_size(self) -> int: ...
    @property
    def sigma_accommodation_amplitude(self) -> float: ...
    @property
    def sigma_adaptation_amplitude(self) -> float: ...

class LeakyIntegrator:
    last_t: float
    rate: float
    scale: float
    value: float
    def __init__(self, scale: float = ..., rate: float = ...) -> None: ...
    def __call__(self, c: float, t: float) -> float: ...

class LeakyIntegratorDecay(Decay):
    def __init__(self, adaptation_amplitude: float = ..., accommodation_amplitude: float = ..., adaptation_rate: float = ..., accommodation_rate: float = ..., sigma_amp: float = ..., sigma_rate: float = ...) -> None: ...
    @property
    def accommodation(self) -> LeakyIntegrator: ...
    @property
    def adaptation(self) -> LeakyIntegrator: ...
    @property
    def sigma_amp(self) -> float: ...
    @property
    def sigma_rate(self) -> float: ...

class Period:
    mu: float
    def __init__(self, arg0: float) -> None: ...
    def tau(self, arg0: float) -> float: ...

class Powerlaw(HistoricalDecay):
    def __init__(self, 
                 adaptation_amplitude: float = ..., 
                 accommodation_amplitude: float = ..., 
                 sigma_adaptation_amplitude: float = ..., 
                 sigma_accommodation_amplitude: float = ..., 
                 offset: float = ..., 
                 exp: float = ...,
                 memory_size: int = ..., 
                 allow_precomputed_accommodation: bool = ..., 
                 cached_decay: bool = ..., cache: list[float] = ...) -> None: ...
    @property
    def exp(self) -> float: ...
    @property
    def offset(self) -> float: ...

class Pulse:
    def __init__(self, arg0: float, arg1: int, arg2: int) -> None: ...
    @property
    def amplitude(self) -> float: ...
    @property
    def electrode(self) -> int: ...
    @property
    def time(self) -> int: ...

class PulseTrain(AbstractPulseTrain):
    def __init__(self, pulse_train: list[list[float]], time_step: float = ..., time_to_ap: float = ..., sigma_ap: float = ...) -> None: ...
    def get_pulse(self, arg0: int) -> Pulse: ...
    @property
    def electrodes(self) -> numpy.ndarray: ...
    @property
    def pulse_times(self) -> numpy.ndarray: ...
    @property
    def pulses(self) -> numpy.ndarray: ...

class RandomGenerator:
    def __init__(self, seed: int, use_random: bool = ...) -> None: ...
    def __call__(self) -> float: ...

class RefractoryPeriod:
    def __init__(self, absolute_refractory_period: float = ..., relative_refractory_period: float = ..., sigma_absolute_refractory_period: float = ..., sigma_relative_refractory_period: float = ...) -> None: ...
    def compute(self, arg0: int, arg1: float, arg2, arg3) -> float: ...
    def randomize(self, arg0) -> RefractoryPeriod: ...
    @property
    def absolute(self) -> Period: ...
    @property
    def relative(self) -> Period: ...
    @property
    def sigma_absolute(self) -> float: ...
    @property
    def sigma_relative(self) -> float: ...

def linspace(arg0: float, arg1: float, arg2: int) -> list[float]: ...
@overload
def phast(i_det: list[float], i_min: list[float], pulse_train: list[list[float]], decay: Decay, relative_spread: float = ..., n_trials: int = ..., refractory_period: RefractoryPeriod = ..., rel, use_random: bool = ..., fiber_id: int = ..., sigma_rs: float = ..., evaluate_in_parallel: bool = ..., time_step: float = ..., time_to_ap: float = ..., store_stats: bool = ...) -> list[FiberStats]: ...
@overload
def phast(fibers: list[Fiber], pulse_train: AbstractPulseTrain, evaluate_in_parallel: bool = ..., n_trials: int = ..., use_random: bool = ...) -> list[FiberStats]: ...

def set_seed(seed: int) -> None: ...



import unittest
import multiprocessing

import numpy as np

import phast


class TestPhast(unittest.TestCase):
    def run_phast(
        self,
        decay: phast.Decay,
        random: bool = True,
        parallel: bool = False,
        store_stats: bool = False,
        n_trials: int = 10,
        duration: float = 0.4,
    ):
        pt = phast.ConstantPulseTrain(duration, 5000, 1e-3, 1e-6)
        phast.set_seed(42)
        fiber = phast.Fiber(
            i_det=[0.000774],
            spatial_constant=[0.866593],
            sigma=[0.000774 * 0.06],
            fiber_id=1200,
            decay=decay,
            store_stats=store_stats,
        )
        return phast.phast([fiber], pt, parallel, n_trials, random)

    def compare_random(self, decay, n_total_spikes, parallel):
        fiber_stats1 = self.run_phast(decay, True, parallel)
        fiber_stats2 = self.run_phast(decay, True, parallel)
        self.assertListEqual(fiber_stats1, fiber_stats2)
        self.assertSetEqual(set(f.n_pulses for f in fiber_stats1), {2000})
        self.assertGreaterEqual(sum(f.n_spikes for f in fiber_stats1), n_total_spikes)
        self.assertNotEqual(len(set(f.n_spikes for f in fiber_stats1)), 1)

    def test_no_random_powerlaw(self):
        decay = phast.Powerlaw()
        fiber_stats = self.run_phast(decay, False, False)
        self.assertSetEqual(set(f.n_spikes for f in fiber_stats), {48})
        self.assertSetEqual(set(f.n_pulses for f in fiber_stats), {2000})

    def test_no_random_exponential(self):
        decay = phast.Exponential()
        fiber_stats = self.run_phast(decay, False, False)
        self.assertSetEqual(set(f.n_spikes for f in fiber_stats), {37})
        self.assertSetEqual(set(f.n_pulses for f in fiber_stats), {2000})

    def test_random_exponential(self):
        self.compare_random(phast.Exponential(), 717, False)

    def test_random_powerlaw(self):
        self.compare_random(phast.Powerlaw(), 842, False)

    def test_random_exponential_parallel(self):
        self.compare_random(phast.Exponential(), 717, True)

    def test_random_powerlaw_parallel(self):
        self.compare_random(phast.Powerlaw(), 842, True)

    def test_fiber_no_random(self):
        fiber = phast.Fiber(
            i_det=[0.000774],
            spatial_constant=[0.866593],
            sigma=[0.000774 * 0.06],
            fiber_id=1200,
            sigma_rs=0.000,
        )
        fiber2 = fiber.randomize()

        self.assertTrue(np.all(np.array(fiber2.sigma) == fiber.sigma))

    def test_fiber_random(self):
        phast.set_seed(42)
        sigma = 1e-4

        fiber = phast.Fiber(
            i_det=[0.000774],
            spatial_constant=[0.866593],
            sigma=[0.000774 * 0.06],
            fiber_id=1200,
            sigma_rs=sigma,
        )
        fiber2 = fiber.randomize()
        self.assertFalse(np.all(np.array(fiber2.sigma) == fiber.sigma))

    def test_rng(self):
        phast.set_seed(42)
        r1 = phast.GENERATOR()
        phast.set_seed(42)
        r2 = phast.GENERATOR()
        self.assertEqual(r1, r2)

    def test_leaky(self):
        decay = phast.LeakyIntegratorDecay(2, 2, 2, 2)
        fiber_stats = self.run_phast(decay, False)
        self.assertSetEqual(set(f.n_spikes for f in fiber_stats), {60})
        self.assertSetEqual(set(f.n_pulses for f in fiber_stats), {2000})

        self.compare_random(decay, 810, True)
        self.compare_random(decay, 810, False)

    def test_store_stats(self):
        decay = phast.LeakyIntegratorDecay(2, 2, 2, 2)
        fiber_stats, *_ = self.run_phast(decay, store_stats=True, n_trials=1)
        stat_names = (
            "accommodation",
            "adaptation",
            "pulse_times",
            "refractoriness",
            "scaled_i_given",
            "stochastic_threshold",
        )
        for name in stat_names:
            self.assertEqual(len(getattr(fiber_stats, name)), 2000)

    def test_dont_store_stats(self):
        decay = phast.LeakyIntegratorDecay(2, 2, 2, 2)
        fiber_stats, *_ = self.run_phast(decay, store_stats=False, n_trials=1)
        stat_names = (
            "accommodation",
            "adaptation",
            "pulse_times",
            "refractoriness",
            "scaled_i_given",
            "stochastic_threshold",
        )
        for name in stat_names:
            self.assertLessEqual(len(getattr(fiber_stats, name)), 1)

    def test_large_pt(self):
        duration = 100.0
        n_trials = multiprocessing.cpu_count()
        decay = phast.LeakyIntegratorDecay()
        fiber_stats = self.run_phast(
            decay, store_stats=False, parallel=True, n_trials=n_trials, duration=duration
        )
        spike_times = phast.spike_times(fiber_stats)
        spike_rate = phast.spike_rate(
            spike_times, num_bins=50, duration=duration, n_trials=n_trials
        )
        self.assertEqual(len(spike_rate), 50)

if __name__ == "__main__":
    unittest.main()



"""PHAST auditory nerve fiber model."""

import os
import warnings
from collections.abc import Iterable
from time import time
from functools import wraps

import numpy as np
import matplotlib.pyplot as plt

from .phastcpp import (
    Decay,
    HistoricalDecay,
    LeakyIntegrator,
    LeakyIntegratorDecay,
    Exponential,
    Fiber,
    FiberStats,
    GENERATOR,
    HistoricalDecay,
    Period,
    Powerlaw,
    Pulse,
    AbstractPulseTrain,
    PulseTrain,
    ConstantPulseTrain,
    RandomGenerator,
    Period,
    RefractoryPeriod,
    phast,
    set_seed,
)

I_DET = os.path.join(os.path.realpath(os.path.dirname(__file__)), "idet.npy")


def generate_stimulus(
    duration: float = 0.4,
    amplitude: float = 0.001,
    rate: int = 5000,
    pw: float = 18e-6,
    time_step: float = None,
    pulse_duration: float = None,
    time_to_ap: float = 0,
    sigma_ap: float = 0,
    mod_depth: float = 0.1,
    mod_start: int = 0,
    mod_freq: float = 100,
    mod_method: str = "litvak2001",
    use_old_pulse_rate: bool = False,
    as_pulse_train: bool = True,
    n_channels: int = 1,
    index: int = 0,
) -> np.ndarray:
    """Generate a (modulated) stimulus.

    Parameters
    ----------
    duration: float = 0.4
        The duration of the stimulus in seconds.
    amplitude: float
        The amplitude of the stimulus in A.
    rate: int = 5000
        The pulse rate in pps.
    pw: float = 18e-6
        The phase width of each pulse.
    time_step: float = None
        The time step increment of the pulse train
    pulse_duration: float = None
        The total duration of the pulse, if not None, this overwrites pw
    time_to_ap: float = 0
        The time until an action potential is observed
    depth: float = 0.1
        Depth of modulation, denotes the relative size of the modulation w.r.t
        the input amplitude.
    mod_start: int = 0
        The index at which to start the modulation
    mod_freq: float = 100
        The frequency used for modulation
    mod_method: str = 'litvak2001'
        The type of modulation used.
    as_pulse_train: bool = False
        Return a PulseTrain object
    n_channels: int
        Passed to wrap_stimulus if as pulse_train == True
    index: int
        Passed to wrap_stimulus if as pulse_train == True
    Returns
    -------
    np.ndarray
        The generated stimulus.

    Notes
    -----
    For the figures from Hu models, the litvak2001 method is used for some reason.

    Pulse duration/width is only used to calculate the time step of the experiments
    """
    if pulse_duration is None:
        pulse_duration = 2 * pw

    pw = pulse_duration / 2

    mus = 1e-6
    if time_step is None:
        time_component = pulse_duration
        if time_to_ap != 0 and time_to_ap < pulse_duration:
            time_component = time_to_ap
        assert time_component > mus, "smallest time step this works for is 1 mu s"
        time_step = np.gcd(int(time_component / mus), int(1 / rate / mus)) * mus
    else:
        duration / time_step
        warnings.warn(
            "Using user-defined time steps, this can lead to rounding errors."
        )

    length = np.floor(duration / time_step).astype(int)
    pt = np.zeros(length)
    pulse_rate = np.floor(1 / rate / time_step).astype(int)

    if use_old_pulse_rate:  # using old pulse rate
        pulse_rate = int((pw / time_step) * np.floor(1 / rate / pw))

    pt[::pulse_rate] = amplitude

    if mod_depth > 0:
        xt = np.arange(length) * time_step  # np.linspace(0, duration, length)
        m_start = np.floor(length / duration * mod_start).astype(int)
        m_sin = np.sin(mod_freq * 2 * np.pi * xt[m_start:])
        if mod_method == "litvak2001":
            pt[m_start:] = (1 - mod_depth) * pt[m_start:] - mod_depth * (
                pt[m_start:] * m_sin
            )
        elif mod_method == "litvak2003a":
            A = pt[m_start:]
            t = xt[: len(A)]
            pt[m_start:] = A * (1 + mod_depth * np.sin(mod_freq * 2 * np.pi * t))
        elif mod_method == "hu":
            pt[m_start:] += mod_depth * pt[m_start:] * m_sin

    pt = wrap_stimulus(pt, n_channels, index)

    if as_pulse_train:
        pt = PulseTrain(pt, time_step, time_to_ap, sigma_ap)

    return pt


def wrap_stimulus(
    stimulus: np.ndarray,
    n_channels: int = 8,
    index: int = 7,
) -> np.ndarray:
    """Wrap a single stimulus in a pulse train of size n_channels at index.

    Parameters
    ----------
    n_channels: int
        The number of electrodes/channels in the pulse train
    indx: int
        The index at which to store the stimulus

    Returns
    -------
        np.ndarray
    """

    pulse_train = np.zeros((n_channels, stimulus.size))
    pulse_train[index, :] = stimulus
    return pulse_train


def spike_times(fiber_stats, time_step=18e-6):
    if not isinstance(fiber_stats, Iterable):
        return fiber_stats.spikes * time_step
    return np.hstack([fs.spikes * time_step for fs in fiber_stats])


def permute_spike_times(spike_times, time_step=18e-6):
    """Permute the spike rates, since they now are only extactly at the time steps
    of the model.

    Parameters
    ----------
    spike_times: np.ndarray
    time_step: float

    Returns
    -------
    np.ndarray
    """
    return np.maximum(np.random.normal(spike_times, time_step), spike_times)


def spike_rate(spike_times, num_bins=None, duration=0.4, binsize=0.05, n_trials=100):
    num_bins = num_bins or int(duration / binsize)
    counts, _ = np.histogram(spike_times, num_bins, (0, duration))
    return counts / n_trials / binsize


def isi(fiber_stats, time_step=18e-6, stack=True):
    def isi_(fs):
        return np.diff(fs.spikes) * time_step

    if isinstance(fiber_stats, Iterable):
        data = [isi_(fs) for fs in fiber_stats]
        if stack:
            data = np.hstack(data)
        return data

    return isi_(fiber_stats)


def plot_spikes(ax: plt.axes, spikes: list, xmax=None, pulse_width=None):
    m = 0
    if pulse_width != None:
        for i, fs in enumerate(spikes):
            ax.scatter(
                fs * pulse_width, np.ones(len(fs)) + i, c="black", s=1, alpha=0.8
            )
        m = pulse_width * (
            xmax + 1
        )  # +1 because of rounding it will not become full pulse train duration
        ax.set_xlim(0, m)
    else:
        for i, fs in enumerate(spikes):
            ax.scatter(fs, np.ones(len(fs)) + i, c="black", s=1, alpha=0.8)
            if any(fs):
                m = max(m, max(fs))
            ax.set_xlim(0, xmax or m)
    ax.set_xlabel("time")
    ax.set_ylabel("trial")


def post_stimulus_time_histogram(ax, spikes: list, pulse_width: float, bin_width=1e-3):
    try:
        n = int(max(map(max, spikes))) + 1
    except ValueError:
        return

    spike_history = np.zeros((len(spikes), n))
    for i, sp in enumerate(spikes):
        spike_history[i, sp] = 1

    sound_duration = pulse_width * (
        n
    )  # this may cause errors if future input is not of exact 0.3 /0.4s length (but instead 0.44s)
    num_bins = int(pulse_width * (n) / bin_width)
    _, spike_times_idx = np.nonzero(spike_history)
    spike_times = spike_times_idx * pulse_width
    bins = np.linspace(0, sound_duration, num_bins)
    ax.hist(
        spike_times, bins
    )  # opmaak: made histtype='stepfilled', rwidth = 2, facecolor = 'k', alpha = 1
    ax.set_xlim(
        0, pulse_width * (n)
    )  # +1 because of rounding it will not become full pulse train duration
    ax.set_ylabel(f"spikes/bin ({bin_width}s)")
    ax.grid()


def plot_pulse_train(pulse_train, figsize=(5, 2)):
    n_channels = min(pulse_train.shape)
    fig, (axes) = plt.subplots(n_channels, 1, sharex=True, sharey=True, figsize=figsize)
    colors = plt.cm.get_cmap("tab20")
    for e, (pulses, ax) in enumerate(zip(pulse_train, axes), 1):
        ax.plot(pulses, c=colors(e))
        ax.grid()
        ax.set_ylabel(e)
    ax.set_xlabel("time")
    plt.tight_layout()


def plot_fiber_stats(fiber_stats):
    if not any(fiber_stats):
        return

    stat_names = ["accommodation", "adaptation", "refractoriness"]
    _, axes = plt.subplots(2, 2, figsize=(15, 8))
    axes = axes.ravel()
    x = fiber_stats[0].pulse_times

    for ax, stat in zip(axes, stat_names):
        data = np.vstack([getattr(fs, stat) for fs in fiber_stats])
        # data[np.isinf(data)] = np.max(data[~np.isinf(data)])
        ax.errorbar(
            x, data.mean(axis=0), yerr=data.std(axis=0), errorevery=100, ecolor="red"
        )
        ax.set_title(stat)
        ax.grid()
    plt.tight_layout()
    stochastic_threshold = np.vstack([fs.stochastic_threshold for fs in fiber_stats])
    axes[-1].hist(stochastic_threshold.ravel(), 1000)
    axes[-1].grid()

    axes[-1].set_title("stochastic threshold")



import os
import platform
from glob import glob
from setuptools import setup

from pybind11.setup_helpers import Pybind11Extension, build_ext

__version__ = "1.0.3"

ext = Pybind11Extension(
    "phast.phastcpp", glob("src/*cpp"), include_dirs=["src"], cxx_std=17
)

if platform.system() in ("Linux", "Darwin"):
    os.environ["CC"] = "g++"
    os.environ["CXX"] = "g++"
    ext._add_cflags(["-O3"])
else:
    ext._add_cflags(["/O2"])


with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    description = f.read()

setup(
    name="phastc",
    author="Jacob de Nobel",
    ext_modules=[ext],
    cmdclass={"build_ext": build_ext},
    description="Phenomological Adaptive STochastic auditory nerve fiber model",
    long_description=description,
    long_description_content_type="text/markdown",
    packages=["phast"],
    package_data={
        "phast": [
            "phast/idet.npy",
        ],
    },
    zip_safe=False,
    version=__version__,
    install_requires=["matplotlib>=3.3.4", "numpy>=1.19.2", "scipy>=1.5.2"],
    include_package_data=True,
)



import os
import unittest
import numpy as np
import phast

DATA_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), "data"))


class TestProblem(unittest.TestCase):
    def setUp(self):
        self.config = {
            "adap_ampl": 0.01,
            "acco_ampl": 0.0003,
            "adap_rate": 2.0,
            "acco_rate": 2.0,
            "arp": 0.4e-3,
            "rrp": 0.8e-3,
            "sigma_arp": 0.1e-3,
            "sigma_rrp": 0.5e-3,
            "rs": 0.06,
            "sigma_rs": 0.04,
        }
        self.pulse_train = np.load(os.path.join(DATA_DIR, "pulse_train_121.npy"))
        self.i_det = np.load(os.path.join(DATA_DIR, "I_det_121.npy"))

    def test_problem(self):
        stimulus = phast.PulseTrain(pulse_train=self.pulse_train, time_step=18e-6)
        fiber_idx = 0
        idet = self.i_det[fiber_idx, :]
        spatial_factor = np.nanmin(self.i_det, axis=0) / idet
        sigma = idet * self.config["rs"]

        refractoriness = phast.RefractoryPeriod(
            absolute_refractory_period=self.config["arp"],
            relative_refractory_period=self.config["rrp"],
            sigma_absolute_refractory_period=self.config["sigma_arp"],
            sigma_relative_refractory_period=self.config["sigma_rrp"],
        )

        decay = phast.LeakyIntegratorDecay(
            adaptation_amplitude=self.config["adap_ampl"],
            accommodation_amplitude=self.config["acco_ampl"],
            adaptation_rate=self.config["adap_rate"],
            accommodation_rate=self.config["acco_rate"],
        )

        fiber = phast.Fiber(
            i_det=idet,
            spatial_constant=spatial_factor,
            sigma=sigma,  # membrane sigma = I_det * RS
            fiber_id=fiber_idx,
            sigma_rs=self.config["sigma_rs"],
            refractory_period=refractoriness,  # object
            decay=decay, 
            store_stats=False,
        )
        fiber_stats = phast.phast([fiber], stimulus)
        spike_times = phast.spike_times(fiber_stats, 18e-6)
        self.assertEqual(len(spike_times), 0)
